--
-- AIVehicle
--
-- Copyright (C) GIANTS Software GmbH, Confidential, All Rights Reserved.

source("dataS/scripts/vehicles/specializations/events/AIVehicleRotateLeftEvent.lua");
source("dataS/scripts/vehicles/specializations/events/AIVehicleRotateRightEvent.lua");
source("dataS/scripts/vehicles/specializations/events/AIVehicleSetStartedEvent.lua");
source("dataS/scripts/vehicles/specializations/events/AIVehicleSetConveyorBeltAngleEvent.lua");

source("dataS/scripts/vehicles/ai/AIDriveStrategy.lua");
source("dataS/scripts/vehicles/ai/AIDriveStrategyCollision.lua");
source("dataS/scripts/vehicles/ai/AIDriveStrategyCombine.lua");
source("dataS/scripts/vehicles/ai/AIDriveStrategyStraight.lua");
source("dataS/scripts/vehicles/ai/AIDriveStrategyConveyor.lua");

source("dataS/scripts/vehicles/ai/AITurnStrategy.lua");
source("dataS/scripts/vehicles/ai/AITurnStrategyDefault.lua");
source("dataS/scripts/vehicles/ai/AITurnStrategyBulb1.lua");           -- "centered bulb"
source("dataS/scripts/vehicles/ai/AITurnStrategyBulb2.lua");           -- "bulb shifted to side", straight first
source("dataS/scripts/vehicles/ai/AITurnStrategyBulb3.lua");           -- "bulb shifted to side", curve first
source("dataS/scripts/vehicles/ai/AITurnStrategyHalfCircle.lua");

source("dataS/scripts/vehicles/ai/AITurnStrategyDefaultReverse.lua");
source("dataS/scripts/vehicles/ai/AITurnStrategyBulb1Reverse.lua");
source("dataS/scripts/vehicles/ai/AITurnStrategyBulb2Reverse.lua");
source("dataS/scripts/vehicles/ai/AITurnStrategyBulb3Reverse.lua");
source("dataS/scripts/vehicles/ai/AITurnStrategyHalfCircleReverse.lua");

---Specialization for AI vehicles
-- @category Specializations
customAIVehicle = {}

customAIVehicle.NUM_BITS_REASONS = 3

customAIVehicle.STOP_REASON_USER = 1
customAIVehicle.STOP_REASON_REGULAR = 2
customAIVehicle.STOP_REASON_UNKOWN = 3
customAIVehicle.STOP_REASON_OUT_OF_FUEL = 4
customAIVehicle.STOP_REASON_OUT_OF_FILL = 5
customAIVehicle.STOP_REASON_BLOCKED_BY_OBJECT = 6
customAIVehicle.STOP_REASON_GRAINTANK_IS_FULL = 7

customAIVehicle.REASON_TEXT_MAPPING = {}
customAIVehicle.REASON_TEXT_MAPPING[customAIVehicle.STOP_REASON_USER] = "ingameNotification_aiVehicleReasonUser"
customAIVehicle.REASON_TEXT_MAPPING[customAIVehicle.STOP_REASON_REGULAR] = "ingameNotification_aiVehicleReasonRegular"
customAIVehicle.REASON_TEXT_MAPPING[customAIVehicle.STOP_REASON_UNKOWN] = "ingameNotification_aiVehicleReasonUnkown"
customAIVehicle.REASON_TEXT_MAPPING[customAIVehicle.STOP_REASON_OUT_OF_FUEL] = "ingameNotification_aiVehicleReasonOutOfFuel"
customAIVehicle.REASON_TEXT_MAPPING[customAIVehicle.STOP_REASON_OUT_OF_FILL] = "ingameNotification_aiVehicleReasonOutOfFill"
customAIVehicle.REASON_TEXT_MAPPING[customAIVehicle.STOP_REASON_BLOCKED_BY_OBJECT] = "ingameNotification_aiVehicleReasonBlockedByObject"
customAIVehicle.REASON_TEXT_MAPPING[customAIVehicle.STOP_REASON_GRAINTANK_IS_FULL] = "ingameNotification_aiVehicleReasonGrainTankIsFull"

customAIVehicle.numHirablesHired = 0;

customAIVehicle.aiDebugRendering = false;

---Checks if all prerequisite specializations are loaded
-- @param table specializations specializations
-- @return boolean hasPrerequisite true if all prerequisite specializations are loaded
-- @includeCode
function customAIVehicle.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(customSteerable, specializations);
end

---Called on loading
-- @param table savegame savegame
-- @includeCode
function customAIVehicle:load(savegame)

    self.onStartAiVehicle = SpecializationUtil.callSpecializationsFunction("onStartAiVehicle");
    self.onStopAiVehicle = SpecializationUtil.callSpecializationsFunction("onStopAiVehicle");
    self.getAdditionalAIPrice = SpecializationUtil.callSpecializationsFunction("getAdditionalAIPrice");
    self.setAIConveyorBeltAngle = SpecializationUtil.callSpecializationsFunction("setAIConveyorBeltAngle");

    self.getDeactivateOnLeave = Utils.overwrittenFunction(self.getDeactivateOnLeave, customAIVehicle.getDeactivateOnLeave);
    self.canStartAIVehicle = customAIVehicle.canStartAIVehicle;
    self.startAIVehicle = customAIVehicle.startAIVehicle;
    self.stopAIVehicle = customAIVehicle.stopAIVehicle;
    self.getVehicleData = customAIVehicle.getVehicleData;
    self.setDriveStrategies = customAIVehicle.setDriveStrategies;


    self.aiVehicleDirectionNode = self.steeringCenterNode;  -- defined and created by ackermann steering
    if self.aiVehicleDirectionNode == nil then
        print("Warning: AIVehicle can't be loaded for "..tostring(self.configFileName)..", because the setup for Ackermann Steering is missing!");
    end

    self.aiIsStarted = false;
    self.isAllowedToDrive = true;

    self.aiImplementList = {};
    self.aiToolsDirtyFlag = true;

    self.driveStrategies = {};

    self.trafficCollisionIgnoreList = {};

    self.didNotMoveTimeout = Utils.getNoNil( getXMLFloat(self.xmlFile, "vehicle.ai.didNotMoveTimeout#value"), 5000);
    if getXMLBool(self.xmlFile, "vehicle.ai.didNotMoveTimeout#deactivated") then
        self.didNotMoveTimeout = math.huge;
    end;

    self.didNotMoveTimer = self.didNotMoveTimeout;

    --
    local aiLightState = Utils.getNoNil( getXMLInt(self.xmlFile, "vehicle.ai.lightState#index"), 3);
    if self.lightStates ~= nil and #self.lightStates > 0 then
        if self.lightStates[aiLightState] == nil then
            aiLightState = 1;
        end
        self.aiLightsTypesMask = 0
        for _, lightType in pairs(self.lightStates[aiLightState]) do
            self.aiLightsTypesMask = bitOR(self.aiLightsTypesMask, 2^lightType);
        end
    end

    -- used for visual debuging
    self.debugTexts = {};
    self.debugLines = {};
    self.debugPoints = {};

    self.pricePerMS = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.ai.pricePerHour"), 2000)/60/60/1000;

    self.isConveyorBelt = hasXMLProperty(self.xmlFile, "vehicle.ai.conveyorBelt");
    if self.isConveyorBelt then
        self.aiConveyorBelt = {};
        self.aiConveyorBelt.minAngle = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.ai.conveyorBelt#minAngle"), 5);
        self.aiConveyorBelt.maxAngle = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.ai.conveyorBelt#maxAngle"), 45);
        self.aiConveyorBelt.stepSize = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.ai.conveyorBelt#stepSize"), 5);
        self.aiConveyorBelt.currentAngle = self.aiConveyorBelt.minAngle;

        self.aiConveyorBelt.speed = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.ai.conveyorBelt#speed"), 1);
        self.aiConveyorBelt.centerWheelIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.ai.conveyorBelt#centerWheelIndex"), 1);
        self.aiConveyorBelt.backWheelIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.ai.conveyorBelt#backWheelIndex"), 2);
    end;

    self.isHired = false;
    self.isHirableBlocked = false;
end

---Called after loading
-- @param table savegame savegame
-- @includeCode
function customAIVehicle:postLoad(savegame)
    if self.isConveyorBelt then
        if savegame ~= nil and not savegame.resetVehicles then
            self.aiConveyorBelt.currentAngle = Utils.getNoNil(getXMLFloat(savegame.xmlFile, savegame.key.."#currentAngle"), self.aiConveyorBelt.minAngle);
        end
    end;
end

---Called on deleting
-- @includeCode
function customAIVehicle:delete()
    if self.aiIsStarted then
        self:stopAIVehicle(customAIVehicle.STOP_REASON_REGULAR);
    end
end

---Returns attributes and nodes to save
-- @param table nodeIdent node ident
-- @return string attributes attributes
-- @return string nodes nodes
-- @includeCode
function customAIVehicle:getSaveAttributesAndNodes(nodeIdent)
    local attributes = "";
    local nodes = "";

    if self.isConveyorBelt then
        attributes = 'currentAngle="'..self.aiConveyorBelt.currentAngle..'"';
    end;

    return attributes, nodes;
end;

---Called on client side on join
-- @param integer streamId streamId
-- @param integer connection connection
-- @includeCode
function customAIVehicle:readStream(streamId, connection)
    local isHired = streamReadBool(streamId);
    if isHired then
        local helperIndex = streamReadUInt8(streamId)
        self:startAIVehicle(helperIndex, true)
    end

    if self.isConveyorBelt then
        self:setAIConveyorBeltAngle(streamReadInt8(streamId), true);
    end;
end

---Called on server side on join
-- @param integer streamId streamId
-- @param integer connection connection
-- @includeCode
function customAIVehicle:writeStream(streamId, connection)
    if streamWriteBool(streamId, self.isHired) then
        streamWriteUInt8(streamId, self.currentHelper.index)
    end

    if self.isConveyorBelt then
        streamWriteInt8(streamId, self.aiConveyorBelt.currentAngle);
    end;
end

function customAIVehicle:mouseEvent(posX, posY, isDown, isUp, button)
end

function customAIVehicle:keyEvent(unicode, sym, modifier, isDown)
end

---Called on update
-- @param float dt time since last call in ms
-- @includeCode
function customAIVehicle:update(dt)

    local activeForInput = not g_gui:getIsGuiVisible() and not g_currentMission.isPlayerFrozen and self.isEntered;

    if activeForInput and customAIVehicle.aiDebugRendering then
        if #self.debugTexts > 0 then
            for i,text in pairs(self.debugTexts) do
                renderText(0.7, 0.92-(0.02*i), 0.02, text);
            end
        end
        if #self.debugLines > 0 then
            for _,l in pairs(self.debugLines) do
                drawDebugLine(l.s[1],l.s[2],l.s[3], l.c[1],l.c[2],l.c[3], l.e[1],l.e[2],l.e[3], l.c[1],l.c[2],l.c[3]);
            end
        end
    end

    if activeForInput then
        if InputBinding.hasEvent(InputBinding.TOGGLE_AI) and not g_currentMission.inGameMessage:getIsVisible() then
            if g_currentMission:getHasPermission("hireAI") then
                if self.aiIsStarted then
                    self:stopAIVehicle(customAIVehicle.STOP_REASON_USER);
                else
                    if self:canStartAIVehicle() then
                        self:startAIVehicle(nil, false);
                    end
                end
            end
        end
        if self.isConveyorBelt then
            if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA3) and not self.aiIsStarted  then
                local newAngle = self.aiConveyorBelt.currentAngle + self.aiConveyorBelt.stepSize
                if newAngle > self.aiConveyorBelt.maxAngle then
                    newAngle = self.aiConveyorBelt.minAngle;
                end

                self:setAIConveyorBeltAngle(newAngle);
            end
        end;
    end

    if self:getIsActive() then
        if self.aiToolsDirtyFlag == true then
            self.aiToolsDirtyFlag = false;
            self:getVehicleData();
        end
    end

    --# only run on server side?
    if not self.isServer then
        return;
    end

    if self.aiIsStarted then
        if self.driveStrategies ~= nil and #self.driveStrategies > 0 then
            for i=1,#self.driveStrategies do
                local driveStrategy = self.driveStrategies[i];
                driveStrategy:update(dt);
            end
        end
    end

    if self.isHired then
        self.forceIsActive = true;
        self.stopMotorOnLeave = false;
        self.steeringEnabled = false;

        -- check light and turn on dependent on daytime
        if self.aiLightsTypesMask ~= nil then

            local dayMinutes = g_currentMission.environment.dayTime/(1000*60);
            local needLights = (dayMinutes > (19*60) or dayMinutes < (6*60));

            if needLights then
                if self.lightsTypesMask ~= self.aiLightsTypesMask then
                    self:setLightsTypesMask(self.aiLightsTypesMask);
                end
            else
                if self.lightsTypesMask ~= 0 then
                    self:setLightsTypesMask(0);
                end
            end

        end

    end;
end

---Called on update tick
-- @param float dt time since last call in ms
-- @includeCode
function customAIVehicle:updateTick(dt)

    --# only run on server side?
    if not self.isServer then
        return;
    end

    if self.isHired and self.isServer and not self.isHirableBlocked then
        local difficultyMultiplier = g_currentMission.missionInfo.buyPriceMultiplier;
        g_currentMission:addSharedMoney(-dt*difficultyMultiplier*self.pricePerMS, "wagePayment");
        g_currentMission:addMoneyChange(-dt*difficultyMultiplier*self.pricePerMS, FSBaseMission.MONEY_TYPE_AI)
    end;

    self.debugTexts = {};
    self.debugLines = {};

    if self.aiIsStarted then
        if self.driveStrategies ~= nil and #self.driveStrategies > 0 then

            local vX,vY,vZ = getWorldTranslation(self.aiVehicleDirectionNode);

            local tX, tZ, moveForwards, maxSpeed, distanceToStop;
            for i=1,#self.driveStrategies do
                local driveStrategy = self.driveStrategies[i];
                tX, tZ, moveForwards, maxSpeed, distanceToStop = driveStrategy:getDriveData(dt, vX,vY,vZ)
                if tX ~= nil or not self.aiIsStarted then
                    break;
                end
            end

            if tX == nil then
                if self.aiIsStarted then -- check if AI is till active, because it might have been kicked by a strategy
                    self:stopAIVehicle(customAIVehicle.STOP_REASON_REGULAR);
                end
                return;
            end

            local lx, lz = AIVehicleUtil.getDriveDirection(self.aiVehicleDirectionNode, tX, vY, tZ);
            if not moveForwards then
                lx, lz = -lx, -lz;
            end

            local acceleration = 1.0;

            local minimumSpeed = 5;
            local lookAheadDistance = 5;

            local distSpeed = math.max(minimumSpeed, maxSpeed * math.min(1, distanceToStop/lookAheadDistance));
            local speedLimit, _ = self:getSpeedLimit();
            maxSpeed = math.min(maxSpeed, distSpeed, speedLimit);
            maxSpeed = math.min(maxSpeed, self.cruiseControl.speed);

            self.isAllowedToDrive = maxSpeed ~= 0;

            local pX,pY,pZ = worldToLocal(self.aiVehicleDirectionNode, tX,vY,tZ);
            if not moveForwards and self.articulatedAxis ~= nil then
                if self.articulatedAxis.aiRevereserNode ~= nil then
                    pX,pY,pZ = worldToLocal(self.articulatedAxis.aiRevereserNode, tX,vY,tZ);
                end
            end

            local doNotSteer = nil;
            if self.isConveyorBelt then
                doNotSteer = true;
            end;

            AIVehicleUtil.driveToPoint(self, dt, acceleration, self.isAllowedToDrive, moveForwards, pX, pZ, maxSpeed, doNotSteer);

            -- worst case check: did not move but should have moved
            if self.isAllowedToDrive and self:getLastSpeed() < 0.5 then
                self.didNotMoveTimer = self.didNotMoveTimer - dt;
            else
                self.didNotMoveTimer = self.didNotMoveTimeout;
            end
            if self.didNotMoveTimer < 0 then
                self:stopAIVehicle(customAIVehicle.STOP_REASON_BLOCKED_BY_OBJECT);
            end

        end
    end
end

---Called on draw
-- @includeCode
function customAIVehicle:draw()
    if g_currentMission:getHasPermission("hireAI") then
        if self.aiIsStarted then
            g_currentMission:addHelpButtonText(g_i18n:getText("action_dismissEmployee"), InputBinding.TOGGLE_AI, nil, GS_PRIO_HIGH);
        else
            if self:canStartAIVehicle() then
                g_currentMission:addHelpButtonText(g_i18n:getText("action_hireEmployee"), InputBinding.TOGGLE_AI, nil, GS_PRIO_HIGH);
            end
        end
    end

    if self.isConveyorBelt then
        if self:getIsActiveForInput(true) and not self.aiIsStarted then
            g_currentMission:addHelpButtonText(string.format(g_i18n:getText("action_conveyorBeltChangeAngle"), string.format("%.0f", self.aiConveyorBelt.currentAngle)), InputBinding.IMPLEMENT_EXTRA3, nil, GS_PRIO_NORMAL);
        end
    end;
    --
    if not self.isServer then
        return;
    end
    if self.aiIsStarted then
        if self.driveStrategies ~= nil and #self.driveStrategies > 0 then
            for i=1,#self.driveStrategies do
                local driveStrategy = self.driveStrategies[i];
                if driveStrategy.draw ~= nil then
                    driveStrategy:draw();
                end
            end
        end
    end
end

---Called on enter vehicle
-- @param boolean isControlling is player controlling the vehicle
-- @includeCode
function customAIVehicle:onEnter(isControlling)
    if self.mapAIHotspot ~= nil then
        self.mapAIHotspot.enabled = false;
    end
end

---Called on leaving the vehicle
-- @includeCode
function customAIVehicle:onLeave()
    if self.mapAIHotspot ~= nil then
        self.mapAIHotspot.enabled = true;
    end
    if self.isHired and self.vehicleCharacter ~= nil then
        self.vehicleCharacter:setCharacterVisibility(true);
    end
end

---Called on attaching a implement
-- @param table implement implement to attach
-- @includeCode
function customAIVehicle:onAttachImplement(implement)
    self.aiToolsDirtyFlag = true;
end

---Called on detaching a implement
-- @param integer implementIndex index of implement to detach
-- @includeCode
function customAIVehicle:onDetachImplement(implementIndex)
    self.aiToolsDirtyFlag = true;
end

---Sets current ai conveyor belt angle
-- @param integer angle new angle
-- @param boolean noEventSend no event send
-- @includeCode
function customAIVehicle:setAIConveyorBeltAngle(angle, noEventSend)
    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            g_server:broadcastEvent(AIVehicleSetConveyorBeltAngleEvent:new(self, angle), nil, nil, self);
        else
            g_client:getServerConnection():sendEvent(AIVehicleSetConveyorBeltAngleEvent:new(self, angle));
        end
    end

    self.aiConveyorBelt.currentAngle = angle;
end

---Returns true if ai can start
-- @return boolean canStart can start ai
-- @includeCode
function customAIVehicle:canStartAIVehicle()
    -- check if reverse driving is available and used, we do not allow the AI to work when reverse driving is enabled
    if self.isReverseDriving ~= nil then
        if self.isReverseDriving or self.isChangingDirection then
            return false;
        end
    end
    if self.aiVehicleDirectionNode == nil then
        return false;
    end
    if g_currentMission.disableAIVehicle then
        return false;
    end
    if customAIVehicle.numHirablesHired >= g_currentMission.maxNumHirables then
        return false;
    end
    if not self.isMotorStarted then
        return false;
    end
    if self.isConveyorBelt then
        return true;
    end;
    if self.aiImplementList ~= nil and #self.aiImplementList > 0 then
        return true;
    else
        return false;
    end
end

---Starts ai vehicle
-- @param integer helperIndex index of hired helper
-- @param boolean noEventSend no event send
-- @includeCode
function customAIVehicle:startAIVehicle(helperIndex, noEventSend)
    if helperIndex ~= nil then
        self.currentHelper = HelperUtil.helperIndexToDesc[helperIndex]
    else
        self.currentHelper = HelperUtil.getRandomHelper()
    end

    HelperUtil.useHelper(self.currentHelper)

    g_currentMission.missionStats:updateStats("workersHired", 1);

    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            g_server:broadcastEvent(AIVehicleSetStartedEvent:new(self, nil, true, self.currentHelper), nil, nil, self);
        else
            g_client:getServerConnection():sendEvent(AIVehicleSetStartedEvent:new(self, nil, true, self.currentHelper));
        end
    end
    self:onStartAiVehicle();
end

---Stops ai vehicle
-- @param integer reason reason
-- @param boolean noEventSend no event send
-- @includeCode
function customAIVehicle:stopAIVehicle(reason, noEventSend)
    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            g_server:broadcastEvent(AIVehicleSetStartedEvent:new(self, reason, false), nil, nil, self);
        else
            g_client:getServerConnection():sendEvent(AIVehicleSetStartedEvent:new(self, reason, false));
        end
    end

    if reason ~= nil and reason ~= customAIVehicle.STOP_REASON_USER then
        local notificationType = FSBaseMission.INGAME_NOTIFICATION_CRITICAL
        if reason == customAIVehicle.STOP_REASON_REGULAR then
            notificationType = FSBaseMission.INGAME_NOTIFICATION_OK
        end
        g_currentMission:addIngameNotification(notificationType, string.format(g_i18n:getText(customAIVehicle.REASON_TEXT_MAPPING[reason]), self.currentHelper.name))
    end

    HelperUtil.releaseHelper(self.currentHelper)

    g_currentMission.missionStats:updateStats("workersHired", -1);

    self:onStopAiVehicle();
end

---Called on start ai vehicle
-- @includeCode
function customAIVehicle:onStartAiVehicle()
    if not self.aiIsStarted then
        if not self.isHired then
            customAIVehicle.numHirablesHired = customAIVehicle.numHirablesHired + 1;
        end;
        self.isHired = true;
        self.isHirableBlocked = false;
        self.forceIsActive = true;
        self.stopMotorOnLeave = false;
        self.steeringEnabled = false;
        self.disableCharacterOnLeave = false;

        if self.vehicleCharacter ~= nil then
           self.vehicleCharacter:delete();
           self.vehicleCharacter:loadCharacter(self.currentHelper.xmlFilename, getUserRandomizedMpColor(self.currentHelper.name))
           if self.isEntered then
                self.vehicleCharacter:setCharacterVisibility(false)
           end
        end

        local hotspotX, _, hotspotZ = getWorldTranslation(self.rootNode);

        local _, textSize = getNormalizedScreenValues(0, 6);
        local _, textOffsetY = getNormalizedScreenValues(0, 11.5);
        local width, height = getNormalizedScreenValues(15,15)
        self.mapAIHotspot = g_currentMission.ingameMap:createMapHotspot("helper", self.currentHelper.name, nil, getNormalizedUVs({776, 520, 240, 240}), {0.052, 0.1248, 0.672, 1}, hotspotX, hotspotZ, width, height, false, false, true, self.components[1].node, true, MapHotspot.CATEGORY_AI, textSize, textOffsetY, {1, 1, 1, 1}, nil, getNormalizedUVs({776, 520, 240, 240}), Overlay.ALIGN_VERTICAL_MIDDLE, 0.7)
        self.aiIsStarted = true;

        if self.isServer then
            self:getVehicleData();
            self:setDriveStrategies();
        end

        self:aiTurnOn();
        for _,implement in pairs(self.aiImplementList) do
            implement.object:aiTurnOn();
        end
    end
end

---Called on stop ai vehicle
-- @includeCode
function customAIVehicle:onStopAiVehicle()
    if self.aiIsStarted then
        if self.isHired then
            customAIVehicle.numHirablesHired = math.max(customAIVehicle.numHirablesHired - 1, 0);
        end;

        self.isHired = false;

        self.forceIsActive = false;
        self.stopMotorOnLeave = true;
        self.steeringEnabled = true;

        self.disableCharacterOnLeave = true;

        if self.vehicleCharacter ~= nil then
           self.vehicleCharacter:delete();
        end

        if self.isEntered or self.isControlled then
            if self.vehicleCharacter ~= nil then
                self.vehicleCharacter:loadCharacter(PlayerUtil.playerIndexToDesc[self.playerIndex].xmlFilename, self.playerColorIndex)
                self.vehicleCharacter:setCharacterVisibility(not self.isEntered)
            end
        end;
        self.currentHelper = nil

        if self.mapAIHotspot ~= nil then
            g_currentMission.ingameMap:deleteMapHotspot(self.mapAIHotspot);
            self.mapAIHotspot = nil;
        end

        self:setCruiseControlState(Drivable.CRUISECONTROL_STATE_OFF, true);
        if self.isServer then
            WheelsUtil.updateWheelsPhysics(self, 0, self.lastSpeedReal, 0, true, self.requiredDriveMode);

            if not g_currentMission.missionInfo.automaticMotorStartEnabled and not self.isControlled then
                self:stopMotor();
            end

            if self.driveStrategies ~= nil and #self.driveStrategies > 0 then
                for i=#self.driveStrategies,1,-1 do
                    self.driveStrategies[i]:delete();
                    table.remove(self.driveStrategies, i);
                end
                self.driveStrategies = {};
            end
        end

        self:aiTurnOff();
        for _,implement in pairs(self.aiImplementList) do
            if implement.object ~= nil then
                implement.object:aiTurnOff();
            end;
        end

        self.aiIsStarted = false;
    end
end

---Returns additional ai price
-- @return float price additional price
-- @includeCode
function customAIVehicle:getAdditionalAIPrice()
    return 0;
end

---Fills aiImplementList with vehicles to use by ai
-- @includeCode
function customAIVehicle:getVehicleData()

    self.aiImplementList = {};

    if self.aiLeftMarker ~= nil and self.aiRightMarker ~= nil and self.aiBackMarker ~= nil then
        table.insert(self.aiImplementList, {object=self});
    end

    AIVehicleUtil.getImplementList(self, self.aiImplementList);

    --# check type and relative position of implement
    if self.aiImplementList ~= nil then
        for _,implement in pairs(self.aiImplementList) do

            if implement.object.attacherVehicle ~= nil then
                local jointDesc = implement.object.attacherVehicle.attacherJoints[implement.jointDescIndex];
                if jointDesc.rotationNode ~= nil then
                    implement.isTool = true;
                else
                    implement.isTool = false;
                end
            else
                implement.isTool = true;
            end

        end
    end

end

---Set drive strategies depending on the vehicle
-- @includeCode
function customAIVehicle:setDriveStrategies()
    if self.aiImplementList ~= nil and #self.aiImplementList > 0 then

        if self.driveStrategies ~= nil and #self.driveStrategies > 0 then
            for i=#self.driveStrategies,1,-1 do
                self.driveStrategies[i]:delete();
                table.remove(self.driveStrategies, i);
            end
            self.driveStrategies = {};
        end

        local foundCombine = false;
        for _,implement in pairs(self.aiImplementList) do
            if SpecializationUtil.hasSpecialization(Combine, implement.object.specializations) then
                foundCombine = true;
                break;
            end
        end
        foundCombine = foundCombine or SpecializationUtil.hasSpecialization(Combine, self.specializations);
        if foundCombine then
            local driveStrategyCombine = AIDriveStrategyCombine:new();
            driveStrategyCombine:setAIVehicle(self);
            table.insert(self.driveStrategies, driveStrategyCombine);
        end

        local driveStrategyCollision = AIDriveStrategyCollision:new();
        local driveStrategyStraight = AIDriveStrategyStraight:new();

        driveStrategyCollision:setAIVehicle(self);
        driveStrategyStraight:setAIVehicle(self);

        table.insert(self.driveStrategies, driveStrategyCollision);
        table.insert(self.driveStrategies, driveStrategyStraight);
    end

    if self.isConveyorBelt then
        if self.driveStrategies ~= nil and #self.driveStrategies > 0 then
            for i=#self.driveStrategies,1,-1 do
                self.driveStrategies[i]:delete();
                table.remove(self.driveStrategies, i);
            end
            self.driveStrategies = {};
        end

        local aiDriveStrategyConveyor = AIDriveStrategyConveyor:new();

        aiDriveStrategyConveyor:setAIVehicle(self);

        table.insert(self.driveStrategies, aiDriveStrategyConveyor);
    end;
end

---Get deactivate on leaving
-- @return boolean deactivateOnLeave deactivate on leaving
-- @includeCode
function customAIVehicle:getDeactivateOnLeave(superFunc)
    local deactivate = true
    if superFunc ~= nil then
        deactivate = deactivate and superFunc(self)
    end

    return deactivate and not self.isHired
end;

---Returns string with states for game stats xml file
-- @return string attributes stats attributes
-- @includeCode
function customAIVehicle:getXMLStatsAttributes()
    if self.isHired then
        return 'isHired="true"';
    end
    return nil;
end

---Turns ai to the right
-- @param boolean force force rotation
-- @includeCode
function customAIVehicle.aiRotateRight(self, force)
    if self.isServer then
        g_server:broadcastEvent(AIVehicleRotateRightEvent:new(self), nil, nil, self);
    end
    if self.onAiRotateRight ~= nil then
        self:onAiRotateRight(force);
    end
    for _,implement in pairs(self.aiImplementList) do
        if implement.object ~= nil then
            implement.object:onAiRotateRight(force);
        end
    end
end

---Turns ai to the left
-- @param boolean force force rotation
-- @includeCode
function customAIVehicle.aiRotateLeft(self, force)
    if self.isServer then
        g_server:broadcastEvent(AIVehicleRotateLeftEvent:new(self), nil, nil, self);
    end
    if self.onAiRotateLeft ~= nil then
        self:onAiRotateLeft(force);
    end
    for _,implement in pairs(self.aiImplementList) do
        if implement.object ~= nil then
            implement.object:onAiRotateLeft(force);
        end
    end
end

---Activates ai debug rendering
-- @includeCode
function customAIVehicle.consoleCommandToggleDebugRenderingAI(unusedSelf)
    customAIVehicle.aiDebugRendering = not customAIVehicle.aiDebugRendering;
end

addConsoleCommand("gsVehicleToggleDebugRenderingAI", "Toggles the debug rendering of the AI vehicles", "customAIVehicle.consoleCommandToggleDebugRenderingAI", nil);
