--
-- Honk
--
-- @author  Manuel Leithner
-- @date  19/12/2013
--
-- Copyright (C) GIANTS Software GmbH, Confidential, All Rights Reserved.

source("dataS/scripts/vehicles/specializations/events/HonkEvent.lua");

---Specialization for honking
-- @category Specializations
customHonk = {};

---Checks if all prerequisite specializations are loaded
-- @param table specializations specializations
-- @return boolean hasPrerequisite true if all prerequisite specializations are loaded
-- @includeCode
function customHonk.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(customDrivable, specializations);
end;

---Called on loading
-- @param table savegame savegame
-- @includeCode
function customHonk:load(savegame)

    self.playHonk = Utils.overwrittenFunction(self.playHonk, customHonk.playHonk);

    if self.isClient then
        self.sampleHonk = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.honkSound", nil, self.baseDirectory, self.components[1].node);
    end;
end;

---Called on deleting
-- @includeCode
function customHonk:delete()
    if self.isClient then
        SoundUtil.deleteSample(self.sampleHonk);
    end;
end;

function customHonk:readStream(streamId, connection)
end;

function customHonk:writeStream(streamId, connection)
end;

function customHonk:mouseEvent(posX, posY, isDown, isUp, button)
end;

function customHonk:keyEvent(unicode, sym, modifier, isDown)
end;

---Called on update
-- @param float dt time since last call in ms
-- @includeCode
function customHonk:update(dt)
    if self.isEntered and self.isClient and self:getIsActiveForInput(false) then
        if self.sampleHonk ~= nil then
            -- only send event if honking key is pressed or released
            if InputBinding.isPressed(InputBinding.HONK) then
                if not self.sampleHonk.isPlaying then
                    self:playHonk(true);
                end;
            else
                if self.sampleHonk.isPlaying then
                    self:playHonk(false);
                end;
            end;
        end;
    end;
end;

function customHonk:updateTick(dt)
end;

---Called on draw
-- @includeCode
function customHonk:draw()
    if self.isClient and self:getIsActiveForInput(true) then
        g_currentMission:addHelpButtonText(g_i18n:getText("action_honk"), InputBinding.HONK, nil, GS_PRIO_VERY_LOW);
    end;
end;

---Called on deactivating sounds
-- @includeCode
function customHonk:onDeactivateSounds()
    if self.isClient then
        SoundUtil.stopSample(self.sampleHonk, true);
        SoundUtil.stop3DSample(self.sampleHonk);
    end;
end;

---Play honk
-- @param boolean isPlaying new is playing
-- @includeCode
function customHonk:playHonk(superFunc, isPlaying, noEventSend)
    if superFunc ~= nil then
        superFunc(self, isPlaying, noEventSend)
    end
    if self.sampleHonk ~= nil then
        HonkEvent.sendEvent(self, isPlaying, noEventSend);

        if isPlaying then
            if self:getIsActive() then
                if self:getIsActiveForSound() then
                    SoundUtil.playSample(self.sampleHonk, 0, 0, nil);
                elseif self.isControlled then
                    -- enable 3d sound to vehicle on another player's game
                    SoundUtil.play3DSample(self.sampleHonk);
                end;
            end;
        else
            customHonk.onDeactivateSounds(self);
        end;
    end;
end;
