source("dataS/scripts/vehicles/specializations/events/VehicleSetBeaconLightEvent.lua");
source("dataS/scripts/vehicles/specializations/events/VehicleSetTurnLightEvent.lua");
source("dataS/scripts/vehicles/specializations/events/VehicleSetLightEvent.lua");

customLights = {};

customLights.TURNLIGHT_OFF = 0;
customLights.TURNLIGHT_LEFT = 1;
customLights.TURNLIGHT_RIGHT = 2;
customLights.TURNLIGHT_HAZARD = 3;
customLights.turnLightSendNumBits = 2;

function customLights.prerequisitesPresent(specializations)
  return true;
end;

function customLights:preLoad(savegame)
  self.getDeactivateLights = Utils.overwrittenFunction(self.getDeactivateLights, customLights.getDeactivateLights);
  self.canToggleLight = Utils.overwrittenFunction(self.canToggleLight, customLights.canToggleLight);
  self.setLightsTypesMask = Utils.overwrittenFunction(self.setLightsTypesMask, customLights.setLightsTypesMask);
  self.setTurnLightState = Utils.overwrittenFunction(self.setTurnLightState, customLights.setTurnLightState);
  self.getIsActiveForLights = Utils.overwrittenFunction(self.getIsActiveForLights, customLights.getIsActiveForLights);
  self.setBrakeLightsVisibility = Utils.overwrittenFunction(self.setBrakeLightsVisibility, customLights.setBrakeLightsVisibility);
  self.setBeaconLightsVisibility = Utils.overwrittenFunction(self.setBeaconLightsVisibility, customLights.setBeaconLightsVisibility);
  self.setReverseLightsVisibility = Utils.overwrittenFunction(self.setReverseLightsVisibility, customLights.setReverseLightsVisibility);
  self.getUseHighProfile = Utils.overwrittenFunction(self.getUseHighProfile, customLights.getUseHighProfile);
  self.onTurnLightStateChanged = SpecializationUtil.callSpecializationsFunction("onTurnLightStateChanged");
  self.onBrakeLightsVisibilityChanged = SpecializationUtil.callSpecializationsFunction("onBrakeLightsVisibilityChanged");
  self.onReverseLightsVisibilityChanged = SpecializationUtil.callSpecializationsFunction("onReverseLightsVisibilityChanged");
  self.onLightsTypesMaskChanged = SpecializationUtil.callSpecializationsFunction("onLightsTypesMaskChanged");
  self.onBeaconLightsVisibilityChanged = SpecializationUtil.callSpecializationsFunction("onBeaconLightsVisibilityChanged");
  self.onDeactivateLights = SpecializationUtil.callSpecializationsFunction("onDeactivateLights");
end;

function customLights.loadLights(self, xmlFile, key)
  local realLight = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#realLight"));
  local decoration = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#decoration"));
  local shaderNode = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#shaderNode"));
  local intensity = getXMLFloat(xmlFile, key .. "#intensity");
  local defaultColor = nil;
  if realLight ~= nil then
    setVisibility(realLight, false);
    defaultColor = {getLightColor(realLight)};
  end
  if decoration ~= nil then
    setVisibility(decoration, false);
  end
  if shaderNode ~= nil then
    setShaderParameter(shaderNode, "lightControl", 0, 0, 0, 0, false);
  end
  return {realLight = realLight, decoration = decoration, shaderNode = shaderNode, isActive = false, defaultColor = defaultColor, intensity = intensity};
end

function customLights.loadLightSetup(self, xmlFile, key, lights, realLightToLight)
  local i = 0;
  local numLightTypes = 0;
  while true do
    local lightKey = string.format(key .. ".light(%d)", i);
    if not hasXMLProperty(xmlFile, lightKey) then
      break;
    end;
    local light = customLights.loadLights(self, xmlFile, lightKey)
    local enableDirection = getXMLInt(xmlFile, lightKey .. "#enableDirection");
    local lightTypes = {Utils.getVectorFromString(getXMLString(xmlFile, lightKey .. "#lightTypes"))};
    local excludedLightTypes = {Utils.getVectorFromString(getXMLString(xmlFile, lightKey.."#excludedLightTypes"))};
    if #lightTypes >= 0 then
      if light.realLight ~= nil or light.decoration ~= nil or light.shaderNode ~= nil then
        if light.realLight ~= nil then
          realLightToLight[light.realLight] = light;
        end
        for i = #lightTypes, 1, -1 do
          local lightType = lightTypes[i]
          if lightType >= 0 and lightType <= 30 then
            numLightTypes = math.max(numLightTypes, lightType+1);
          else
            table.remove(lightTypes, i);
            print("Warning: lightType must be between 0 and 30 for light '" ..lightKey.."' in '"..self.configFileName.."'");
          end
        end
        light.excludedLightTypes = excludedLightTypes;
        light.enableDirection = enableDirection;
        light.lightTypes = lightTypes;
        table.insert(lights, light);
      else
        print("Warning: decoration or/and realLight or/and shaderNode missing for light '" ..lightKey.."' in '"..self.configFileName.."'")
      end
    else
      print("Warning: lightType missing for light '" ..lightKey.."' in '"..self.configFileName.."'");
    end;
    i = i + 1;
  end;
  return numLightTypes;
end

function customLights:load(savegame)
  if hasXMLProperty(self.xmlFile, "vehicle.lights.light1") then
    print("Warning: lights.light in '".. self.configFileName .."' is not supported anymore. Use lights.light#realLight instead.");
  end
  if hasXMLProperty(self.xmlFile, "vehicle.lightCoronas.lightCorona") then
    print("Warning: lightCoronas.lightCorona in '".. self.configFileName .."' is not supported anymore. Use lights.light#decoration instead.");
  end
  if hasXMLProperty(self.xmlFile, "vehicle.lightCones.lightCone") then
    print("Warning: lightCones.lightCone in '".. self.configFileName .."' is not supported anymore.");
  end
  if getXMLString(self.xmlFile, "vehicle.brakeLights.brakeLight(0)#index") ~= nil then
    print("Warning: brakeLights.brakeLight#index in '".. self.configFileName .."' is not supported anymore. Use brakeLights.brakeLight#realLight and brakeLights.brakeLight#decoration instead.");
  end
  if getXMLString(self.xmlFile, "vehicle.reverseLights.reverseLight(0)#index") ~= nil then
    print("Warning: reverseLights.reverseLight#index in '".. self.configFileName .."' is not supported anymore. Use reverseLights.reverseLight#realLight and reverseLights.reverseLight#decoration instead.");
  end
  if getXMLString(self.xmlFile, "vehicle.turnSignals.turnSignalLeft(0)#index") ~= nil then
    print("Warning: turnSignals.turnSignalLeft#index in '".. self.configFileName .."' is not supported anymore. Use turnLights.turnLightLeft#realLight and turnLights.turnLightLeft#decoration instead.");
  end
  if getXMLString(self.xmlFile, "vehicle.turnSignals.turnSignalRight(0)#index") ~= nil then
    print("Warning: turnSignals.turnSignalRight#index in '".. self.configFileName .."' is not supported anymore. Use turnLights.turnLightRight#realLight and turnLights.turnLightRight#decoration instead.");
  end
  self.maxNumRealLights = g_maxNumRealVehicleLights;
  self.lights = {};
  self.lights.low = {};
  self.lights.high = {}
  self.numLightTypes = 0;
  local realLightToLight = {};
  local numLowLightTypes = customLights.loadLightSetup(self, self.xmlFile, "vehicle.lights.low", self.lights.low, realLightToLight);
  local numHighLightTypes = customLights.loadLightSetup(self, self.xmlFile, "vehicle.lights.high", self.lights.high, realLightToLight);
  if numLowLightTypes == numHighLightTypes then
    self.numLightTypes = numHighLightTypes;
  else
    print("Warning: Low- and high-profile light setups are different. Both light setups must have equal number of lightTypes.")
  end
  self.lightStates = {};
  local i = 0;
  while true do
    local key = string.format("vehicle.lights.states.state(%d)", i);
    if not hasXMLProperty(self.xmlFile, key) then
      break;
    end;
    local lightTypes = {Utils.getVectorFromString(getXMLString(self.xmlFile, key.."#lightTypes"))};
    table.insert(self.lightStates, lightTypes);
    i = i + 1;
  end
  self.brakeLights = {};
  local i = 0;
  while true do
    local key = string.format("vehicle.brakeLights.brakeLight(%d)", i);
    if not hasXMLProperty(self.xmlFile, key) then
      break;
    end;
    local light = customLights.loadLights(self, self.xmlFile, key)
    if light.realLight ~= nil or light.decoration ~= nil then
      if light.realLight ~= nil and realLightToLight[light.realLight] ~= nil then
        light.backLight = realLightToLight[light.realLight];
        realLightToLight[light.realLight].brakeLight = light;
      end
      table.insert(self.brakeLights, light);
    end;
    i = i + 1;
  end;
  self.brakeLightsVisibility = false;
  self.reverseLights = {};
  local i = 0;
  while true do
    local key = string.format("vehicle.reverseLights.reverseLight(%d)", i);
    if not hasXMLProperty(self.xmlFile, key) then
      break;
    end;
    local light = customLights.loadLights(self, self.xmlFile, key)
    if light.realLight ~= nil or light.decoration ~= nil then
      table.insert(self.reverseLights, light);
    end;
    i = i + 1;
  end;
  self.reverseLightsVisibility = false;
  self.turnLights = {};
  self.turnLights.left = {};
  self.turnLights.right = {};
  local i = 0;
  while true do
    local keyLeft = string.format("vehicle.turnLights.turnLightLeft(%d)", i);
    local keyRight = string.format("vehicle.turnLights.turnLightRight(%d)", i);
    if not hasXMLProperty(self.xmlFile, keyLeft) and not hasXMLProperty(self.xmlFile, keyRight) then
      break;
    end;
    if hasXMLProperty(self.xmlFile, keyLeft) then
      local light = customLights.loadLights(self, self.xmlFile, keyLeft)
      if light.realLight ~= nil or light.decoration ~= nil then
        if light.realLight ~= nil then
          light.defaultColor = {getLightColor(light.realLight)};
        end
        table.insert(self.turnLights.left, light);
      end;
    end;
    if hasXMLProperty(self.xmlFile, keyRight) then
      local light = customLights.loadLights(self, self.xmlFile, keyRight)
      if light.realLight ~= nil or light.decoration ~= nil then
        if light.realLight ~= nil then
          light.defaultColor = {getLightColor(light.realLight)};
        end
        table.insert(self.turnLights.right, light);
      end;
    end;
    i = i + 1;
  end;
  self.turnLightState = customLights.TURNLIGHT_OFF;
  self.hasTurnLights = #self.turnLights.left > 0 or #self.turnLights.right > 0
  local realBeaconLights = g_gameSettings:getValue("realBeaconLights")
  self.beaconLights = {};
  local i = 0;
  while true do
    local key = string.format("vehicle.beaconLights.beaconLight(%d)", i);
    if not hasXMLProperty(self.xmlFile, key) then
      break;
    end;
    local light = customLights.loadLights(self, self.xmlFile, key);
    local node = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#index"));
    light.rotationNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#rotationNode"));
    local filename = getXMLString(self.xmlFile, key.."#filename");
    if filename ~= nil then
      local i3dNode = Utils.loadSharedI3DFile(filename, self.baseDirectory, false, false, false);
      if i3dNode ~= 0 then
        light.filename = filename;
        local rootNode = getChildAt(i3dNode, 0);
        link(node, rootNode);
        delete(i3dNode);
        setTranslation(rootNode, 0,0,0);
        local rotationIndex = getUserAttribute(rootNode, "rotation");
        if rotationIndex ~= nil then
          light.rotationNode = Utils.indexToObject(rootNode, rotationIndex);
        end;
        local decorationIndex = getUserAttribute(rootNode, "decoration");
        if decorationIndex ~= nil then
          light.decoration = Utils.indexToObject(rootNode, decorationIndex);
        end;
        local realLightIndex = getUserAttribute(rootNode, "realLight");
        if realLightIndex ~= nil then
          light.realLight = Utils.indexToObject(rootNode, realLightIndex);
          if light.realLight ~= nil then
            if realBeaconLights then
              light.defaultColor = {getLightColor(light.realLight)};
            else
              setVisibility(light.realLight, false)
            end
          end;
        end;
      end
    end;
    if node ~= nil then
      light.speed = Utils.getNoNil(getXMLFloat(self.xmlFile, key.."#speed"), 0.02);
      if light.speed > 0 then
        local rot = math.random(0, math.pi*2);
        if light.rotationNode ~= nil then
          setRotation(light.rotationNode, 0, rot, 0);
        end;
      end;
      if light.decoration ~= nil then
        setVisibility(light.decoration, false);
      end
      if light.shaderNode ~= nil then
        setShaderParameter(light.shaderNode, "lightControl", 0, 0, 0, 0, false);
      end
      if light.realLight ~= nil then
        setVisibility(light.realLight, false);
      end
      table.insert(self.beaconLights, light);
    end;
    i = i + 1;
  end;
  self.beaconLightsActive = false;
  self.lightsTypesMask = 0;
end;

function customLights:delete()
  for _, beaconLight in pairs(self.beaconLights) do
    if beaconLight.filename ~= nil then
      Utils.releaseSharedI3DFile(beaconLight.filename, self.baseDirectory, true);
    end;
  end;
end;

function customLights:readStream(streamId, connection)
  local lightsTypesMask = streamReadInt32(streamId);
  self:setLightsTypesMask(lightsTypesMask, true, true);
  local beaconLightsActive = streamReadBool(streamId);
  self:setBeaconLightsVisibility(beaconLightsActive, true, true);
  local turnLightState = streamReadUIntN(streamId, customLights.turnLightSendNumBits);
  self:setTurnLightState(turnLightState, true, true);
end;

function customLights:writeStream(streamId, connection)
  streamWriteInt32(streamId, self.lightsTypesMask);
  streamWriteBool(streamId, self.beaconLightsActive);
  streamWriteUIntN(streamId, self.turnLightState, customLights.turnLightSendNumBits);
end;

function customLights:mouseEvent(posX, posY, isDown, isUp, button)
end;

function customLights:keyEvent(unicode, sym, modifier, isDown)
end;

function customLights:update(dt)
  if self.beaconLightsActive then
    for _, beaconLight in pairs(self.beaconLights) do
      if beaconLight.rotationNode ~= nil then
        rotate(beaconLight.rotationNode, 0, beaconLight.speed*dt, 0);
      end
    end;
  end;
  if self.turnLightState ~= customLights.TURNLIGHT_OFF then
    if self:getUseHighProfile() then
      local alpha = Utils.clamp((math.cos(7*getShaderTimeSec()) + 0.2), 0, 1);
      if self.turnLightState == customLights.TURNLIGHT_LEFT or self.turnLightState == customLights.TURNLIGHT_HAZARD then
        for _, signal in pairs(self.turnLights.left) do
          if signal.realLight ~= nil then
            setLightColor(signal.realLight, signal.defaultColor[1]*alpha, signal.defaultColor[2]*alpha, signal.defaultColor[3]*alpha);
            for i=0, getNumOfChildren(signal.realLight)-1 do
              setLightColor(getChildAt(signal.realLight, i), signal.defaultColor[1]*alpha, signal.defaultColor[2]*alpha, signal.defaultColor[3]*alpha);
            end
          end
        end;
      end
      if self.turnLightState == customLights.TURNLIGHT_RIGHT or self.turnLightState == customLights.TURNLIGHT_HAZARD then
        for _, signal in pairs(self.turnLights.right) do
          if signal.realLight ~= nil then
            setLightColor(signal.realLight, signal.defaultColor[1]*alpha, signal.defaultColor[2]*alpha, signal.defaultColor[3]*alpha);
            for i = 0, getNumOfChildren(signal.realLight) - 1 do
              setLightColor(getChildAt(signal.realLight, i), signal.defaultColor[1]*alpha, signal.defaultColor[2]*alpha, signal.defaultColor[3]*alpha);
            end
          end
        end;
      end
    end
  end
end;

function customLights:draw()
end;

function customLights:getIsActiveForLights(superFunc)
  if superFunc ~= nil then
    if not superFunc(self) then
      return false;
    end
  end
  if self.isEntered and self:canToggleLight() then
    return true;
  end;
  if self.attacherVehicle ~= nil and (self.isSteerable == nil or self.isSteerable == false) then
    return self.attacherVehicle:getIsActiveForLights();
  end;
  return false;
end;

function customLights:canToggleLight(superFunc)
  if superFunc ~= nil then
    if not superFunc(self) then
      return false;
    end
  end
  if self:getIsHired() then
    return false;
  end
  if self.numLightTypes == 0 then
    return false;
  end;
  if not g_currentMission.controlPlayer and g_currentMission.controlledVehicle == self then
    return true;
  else
    return false;
  end;
end;

function customLights:getUseHighProfile(superFunc)
  if superFunc ~= nil then
    if not superFunc(self) then
      return false;
    end
  end
  local lightsProfile = g_gameSettings:getValue("lightsProfile")
  return lightsProfile == GS_PROFILE_VERY_HIGH or (lightsProfile == GS_PROFILE_HIGH and self:getIsActiveForLights())
end

function customLights:setLightsTypesMask(superFunc, lightsTypesMask, force, noEventSend)
  if superFunc ~= nil then
    if not superFunc(self, lightsTypesMask, noEventSend) then
      return false;
    end
  end
  if lightsTypesMask ~= self.lightsTypesMask or force then
    if noEventSend == nil or noEventSend == false then
      if g_server ~= nil then
        g_server:broadcastEvent(VehicleSetLightEvent:new(self, lightsTypesMask), nil, nil, self);
      else
        g_client:getServerConnection():sendEvent(VehicleSetLightEvent:new(self, lightsTypesMask));
      end;
    end;
    self.lightsTypesMask = lightsTypesMask;
    local activeLightSetup = self.lights.low;
    local inactiveLightSetup = self.lights.high;
    if self:getUseHighProfile() then
      activeLightSetup = self.lights.high;
      inactiveLightSetup = self.lights.low;
    end
    for _, light in pairs(inactiveLightSetup) do
      if light.decoration ~= nil then
        setVisibility(light.decoration, false);
      end
      if light.shaderNode ~= nil then
        setShaderParameter(light.shaderNode, "lightControl", 0, 0, 0, 0, false);
      end
      if light.realLight ~= nil then
        setVisibility(light.realLight, false);
      end
    end
    for _,light in pairs(activeLightSetup) do
      local lightActive = false;
      for _, lightType in pairs(light.lightTypes) do
        if bitAND(lightsTypesMask, 2^lightType) ~= 0 then
          lightActive = true;
          break;
        end
      end
      if light.enableDirection ~= nil then
        lightActive = lightActive and light.enableDirection == self.reverserDirection;
      end
      if light.decoration ~= nil then
        setVisibility(light.decoration, lightActive);
      end
      if light.shaderNode ~= nil then
        if lightActive then
          setShaderParameter(light.shaderNode, "lightControl", light.intensity, 0, 0, 0, false);
        else
          setShaderParameter(light.shaderNode, "lightControl", 0, 0, 0, 0, false);
        end
      end
      if light.realLight ~= nil then
        if lightActive then
          for _, excludedLightType in pairs(light.excludedLightTypes) do
            if bitAND(lightsTypesMask, 2^excludedLightType) ~= 0 then
              lightActive = false;
              break;
            end
          end
        end
        if lightActive then
          setVisibility(light.realLight, true)
        else
          local active = false;
          if light.brakeLight ~= nil then
            if light.brakeLight.isActive then
              active = true;
            end
          end
          setVisibility(light.realLight, active)
        end
      end
      light.isActive = lightActive;
    end
    for _,v in pairs(self.attachedImplements) do
      if v.object ~= nil then
        v.object:setLightsTypesMask(lightsTypesMask, true, true);
      end
    end;
    self:onLightsTypesMaskChanged(lightsTypesMask);
  end;
  return true;
end;

function customLights:setBeaconLightsVisibility(superFunc, visibility, force, noEventSend)
  if superFunc ~= nil then
    if not superFunc(self, visibility, noEventSend) then
      return false;
    end;
  end
  if visibility ~= self.beaconLightsActive or force then
    if noEventSend == nil or noEventSend == false then
      if g_server ~= nil then
        g_server:broadcastEvent(VehicleSetBeaconLightEvent:new(self, visibility), nil, nil, self);
      else
        g_client:getServerConnection():sendEvent(VehicleSetBeaconLightEvent:new(self, visibility));
      end;
    end;
    self.beaconLightsActive = visibility;
    local realBeaconLights = g_gameSettings:getValue("realBeaconLights")
    for _, beaconLight in pairs(self.beaconLights) do
      if realBeaconLights and beaconLight.realLight ~= nil then
        setVisibility(beaconLight.realLight, visibility);
      end
      if beaconLight.decoration ~= nil then
        setVisibility(beaconLight.decoration, visibility);
      end
    end;
    for _,v in pairs(self.attachedImplements) do
      if v.object ~= nil then
        v.object:setBeaconLightsVisibility(visibility, true, true);
      end
    end;
    self:onBeaconLightsVisibilityChanged(visibility);
  end;
  return true;
end;

function customLights:setTurnLightState(superFunc, state, force, noEventSend)
  if superFunc ~= nil then
    if not superFunc(self, state, noEventSend) then
      return false;
    end;
  end
  if state ~= self.turnLightState or force then
    if noEventSend == nil or noEventSend == false then
      if g_server ~= nil then
        g_server:broadcastEvent(VehicleSetTurnLightEvent:new(self, state), nil, nil, self);
      else
        g_client:getServerConnection():sendEvent(VehicleSetTurnLightEvent:new(self, state));
      end;
    end;
    local useHighProfile = self:getUseHighProfile()
    for _, signal in pairs(self.turnLights.left) do
      if signal.realLight ~= nil then
        if useHighProfile then
          setVisibility(signal.realLight, state == customLights.TURNLIGHT_LEFT or state == customLights.TURNLIGHT_HAZARD);
        end
      end
      if signal.decoration ~= nil then
        setVisibility(signal.decoration, state == customLights.TURNLIGHT_LEFT or state == customLights.TURNLIGHT_HAZARD);
      end
      if signal.shaderNode ~= nil then
        if state == customLights.TURNLIGHT_LEFT or state == customLights.TURNLIGHT_HAZARD then
          setShaderParameter(signal.shaderNode, "lightControl", signal.intensity, 0, 0, 0, false);
        else
          setShaderParameter(signal.shaderNode, "lightControl", 0, 0, 0, 0, false);
        end
      end
    end;
    for _, signal in pairs(self.turnLights.right) do
      if signal.realLight ~= nil then
        if useHighProfile then
          setVisibility(signal.realLight, state == customLights.TURNLIGHT_RIGHT or state == customLights.TURNLIGHT_HAZARD);
        end
      end
      if signal.decoration ~= nil then
        setVisibility(signal.decoration, state == customLights.TURNLIGHT_RIGHT or state == customLights.TURNLIGHT_HAZARD);
      end
      if signal.shaderNode ~= nil then
        if state == customLights.TURNLIGHT_RIGHT or state == customLights.TURNLIGHT_HAZARD then
          setShaderParameter(signal.shaderNode, "lightControl", signal.intensity, 0, 0, 0, false);
        else
          setShaderParameter(signal.shaderNode, "lightControl", 0, 0, 0, 0, false);
        end
      end
    end;
    for _,v in pairs(self.attachedImplements) do
      if v.object ~= nil then
        v.object:setTurnLightState(state, true, true);
      end
    end;
    self:onTurnLightStateChanged(state);
    self.turnLightState = state;
  end;
  return true;
end;

function customLights:setBrakeLightsVisibility(superFunc, visibility)
  if superFunc ~= nil then
    if not superFunc(self, visibility) then
      return false;
    end;
  end
  if visibility ~= self.brakeLightsVisibility then
    self.brakeLightsVisibility = visibility;
    local useHighProfile = self:getUseHighProfile();
    for _, brakeLight in pairs(self.brakeLights) do
      if brakeLight.realLight ~= nil then
        if useHighProfile then
          brakeLight.isActive = visibility;
          if visibility then
            if brakeLight.backLight ~= nil then
              local color = brakeLight.backLight.defaultColor;
              setLightColor(brakeLight.realLight, color[1]*2, color[2]*2, color[3]*2);
              for i = 0, getNumOfChildren(brakeLight.realLight) - 1 do
                setLightColor(getChildAt(brakeLight.realLight, i), color[1]*2, color[2]*2, color[3]*2);
              end
            end
            setVisibility(brakeLight.realLight, true);
            brakeLight.isActive = true;
          else
            local isVisible = false;
            if brakeLight.backLight ~= nil then
              local color = brakeLight.backLight.defaultColor;
              setLightColor(brakeLight.realLight, color[1], color[2], color[3]);
              for i = 0, getNumOfChildren(brakeLight.realLight) - 1 do
                setLightColor(getChildAt(brakeLight.realLight, i), color[1], color[2], color[3]);
              end
              if brakeLight.backLight.isActive then
                isVisible = true;
              end
            end
            setVisibility(brakeLight.realLight, isVisible);
          end;
          if brakeLight.backLight ~= nil then
          end
        end
      end
      if brakeLight.decoration ~= nil then
        setVisibility(brakeLight.decoration, visibility);
      end   
      if brakeLight.shaderNode ~= nil then
        if visibility then
          setShaderParameter(brakeLight.shaderNode, "lightControl", brakeLight.intensity, 0, 0, 0, false);
        else
          setShaderParameter(brakeLight.shaderNode, "lightControl", 0, 0, 0, 0, false);
        end
      end
    end
    for _,v in pairs(self.attachedImplements) do
      if v.object ~= nil then
        v.object:setBrakeLightsVisibility(visibility);
      end
    end
    self:onBrakeLightsVisibilityChanged(visibility);
  end
  return true;
end

function customLights:setReverseLightsVisibility(superFunc, visibility)
  if superFunc ~= nil then
    if not superFunc(self, visibility) then
      return false;
    end;
  end
  local useHighProfile = self:getUseHighProfile()
  if visibility ~= self.reverseLightsVisibility then
    self.reverseLightsVisibility = visibility;
    for _, reverseLight in pairs(self.reverseLights) do
      if reverseLight.realLight ~= nil then
        if useHighProfile then
          setVisibility(reverseLight.realLight, visibility);
        end
      end
      if reverseLight.decoration ~= nil then
        setVisibility(reverseLight.decoration, visibility);
      end
      if reverseLight.shaderNode ~= nil then
        if visibility then
          setShaderParameter(reverseLight.shaderNode, "lightControl", reverseLight.intensity, 0, 0, 0, false);
        else
          setShaderParameter(reverseLight.shaderNode, "lightControl", 0, 0, 0, 0, false);
        end
      end
    end
    for _,v in pairs(self.attachedImplements) do
      if v.object ~= nil then
        v.object:setReverseLightsVisibility(visibility);
      end
    end
    self:onReverseLightsVisibilityChanged(visibility);
  end
  return true;
end;

function customLights:getDeactivateLights(superFunc)
  if superFunc ~= nil then
    return superFunc(self)
  end
  return true
end

function customLights:onLeave()
  if self:getDeactivateLights() then
    self:onDeactivateLights()
  end
end

function customLights:onDeactivateLights()
  self:setLightsTypesMask(0, true, true);
  self:setBeaconLightsVisibility(false, true, true);
  self:setTurnLightState(customLights.TURNLIGHT_OFF, true, true);
  self:setBrakeLightsVisibility(false);
  self:setReverseLightsVisibility(false);
end;

function customLights:onEnter(isControlling)
  self:setLightsTypesMask(self.lightsTypesMask, true, true);
  self:setBeaconLightsVisibility(self.beaconLightsActive, true, true);
  self:setTurnLightState(self.turnLightState, true, true);
end

function customLights:onStartReverseDirectionChange()
  if self.lightsTypesMask > 0 then
    self:setLightsTypesMask(self.lightsTypesMask, true, true);
  end
end
